#!/usr/bin/env python3

import hilmux
from hilmux_proto import *
import set_HILmux

import dataclasses
import tkinter as tk
from tkinter import ttk
import threading


@dataclasses.dataclass
class HilmuxState:
    serialnumber_text: str
    production_text: Optional[str] = "..."
    commit_text: Optional[str] = "..."
    error_text: str = ""
    uc1_mode: Optional[UcMode] = None
    uc2_mode: Optional[UcMode] = None
    lock_mode: Optional[LockMode] = None
    
    def is_locked(self):
        return self.lock_mode is None or self.lock_mode == LockMode.Locked
    
    def copy(self) -> 'HilmuxState':
        return dataclasses.replace(self)


class DeviceControllerApp:
    def __init__(self, root, discovered_device: hilmux.DiscoveredDevice):
        self.root = root
        self.root.title("HILmux Example GUI")
        self.root.geometry("480x260")
        
        self.hilmux = hilmux.Hilmux(discovered_device.device)
        self.state = HilmuxState("HILmux-" + discovered_device.serialnumber)
        
        self.create_widgets()

        # Start background thread to simulate state updates
        self.background_thread = threading.Thread(target=self.background_thread_target, daemon=True)
        self.background_thread.start()


    def create_widgets(self):
        # Title (Serialnumber)
        title_frame = ttk.Frame(self.root, padding=10)
        title_frame.pack(fill=tk.X, expand=True)
        self.serial_label = ttk.Label(title_frame, text="", font=('TkDefaultFont', 14, 'bold'))
        self.serial_label.pack()

        # Overview (info_production, info_commit)
        overview_frame = ttk.Frame(self.root)
        overview_frame.pack(fill=tk.X, expand=True)
        self.production_label = ttk.Label(overview_frame, text="", font=('TkDefaultFont', 10))
        self.production_label.pack(anchor="w", padx=10)
        self.commit_label = ttk.Label(overview_frame, text="Commit: abcdefgh", font=('TkDefaultFont', 10))
        self.commit_label.pack(anchor="w", padx=10)

        # uCs (uC1 and uC2 radio buttons)
        ucs_frame = ttk.Frame(self.root)
        ucs_frame.pack(padx=10, pady=10, fill='x')
        
        # uC 1
        uc1_frame = ttk.LabelFrame(ucs_frame, text=set_HILmux.config.uc1_display_name if set_HILmux.config is not None else "uC 1")
        uc1_frame.pack(side='left', expand=True, fill='both', padx=5)
        self.uc1_var = tk.StringVar(value="")
        self.uc1_options: dict[UcMode, ttk.Radiobutton] = {}
        for mode in UcMode:
            rb = ttk.Radiobutton(uc1_frame, text=mode.name, variable=self.uc1_var, value=mode.name, command=lambda mode=mode: self.change_uc1(mode))
            rb.pack(anchor='w', pady=2)
            self.uc1_options[mode] = rb
        
        # uC 2
        uc2_frame = ttk.LabelFrame(ucs_frame, text=set_HILmux.config.uc2_display_name if set_HILmux.config is not None else "uC 2")
        uc2_frame.pack(side='left', expand=True, fill='both', padx=5)
        self.uc2_var = tk.StringVar(value="")
        self.uc2_options: dict[UcMode, ttk.Radiobutton] = {}
        for mode in UcMode:
            rb = ttk.Radiobutton(uc2_frame, text=mode.name, variable=self.uc2_var, value=mode.name, command=lambda mode=mode: self.change_uc2(mode))
            rb.pack(anchor='w', pady=2)
            self.uc2_options[mode] = rb

        # Lock/Unlock Checkbox in Top Right
        lock_frame = ttk.Frame(self.root)
        lock_frame.pack(fill=tk.X, expand=True)
        self.lock_var = tk.BooleanVar(value=self.state.is_locked())
        self.lock_checkbox = ttk.Checkbutton(lock_frame, text="Locked", variable=self.lock_var, command=self.toggle_lock)
        self.lock_checkbox.pack()

        # Error Label
        bottom_frame = ttk.Frame(self.root, padding=5)
        bottom_frame.pack(fill=tk.X, expand=True)
        self.error_label = ttk.Label(bottom_frame, text="", foreground="red")
        self.error_label.pack(side="left", fill="x", expand=True)

        self.apply_state()

    
    def toggle_lock(self):
        locked = self.lock_var.get()
        self.lock_var.set(not locked)
        self.hilmux.send(CommandSetLockMode(LockMode.Locked if locked else LockMode.Unlocked))


    def change_uc1(self, mode):
        self.hilmux.send(CommandSetUc1Mode(mode))
        self.apply_state()


    def change_uc2(self, mode):
        self.hilmux.send(CommandSetUc2Mode(mode))
        self.apply_state()


    def apply_state(self):
        self.lock_var.set(self.state.is_locked())
        self.production_label.config(text=self.state.production_text)
        self.commit_label.config(text=self.state.commit_text)
        self.error_label.config(text=self.state.error_text)
        self.serial_label.config(text=self.state.serialnumber_text)
        self.uc1_var.set("" if self.state.uc1_mode is None else self.state.uc1_mode.name)
        self.uc2_var.set("" if self.state.uc2_mode is None else self.state.uc2_mode.name)

        # Update state of Lauterbach radio buttons
        for mode in UcMode:
            if not self.state.is_locked() and mode == UcMode.Lauterbach:
                self.uc1_options[mode].configure(state=tk.DISABLED if self.state.uc2_mode == UcMode.Lauterbach else tk.NORMAL)
                self.uc2_options[mode].configure(state=tk.DISABLED if self.state.uc1_mode == UcMode.Lauterbach else tk.NORMAL)
            else:
                self.uc1_options[mode].configure(state=tk.DISABLED if self.state.is_locked() else tk.NORMAL)
                self.uc2_options[mode].configure(state=tk.DISABLED if self.state.is_locked() else tk.NORMAL)        


    def on_close(self):
        self.hilmux.close()
        self.root.destroy()


    def background_thread_target(self):
        try:
            self.hilmux.send(CommandGetAliveFactorySetting())
            self.hilmux.send(CommandGetUc1Mode())
            self.hilmux.send(CommandGetUc2Mode())
            self.hilmux.send(CommandGetLockMode())
            while True:
                state = self.hilmux.recv()
                if state is None:
                    continue
                print(state)
                if isinstance(state, StatusUc1Mode):
                    self.state.uc1_mode = state.uc_mode
                elif isinstance(state, StatusUc2Mode):
                    self.state.uc2_mode = state.uc_mode
                elif isinstance(state, StatusLockMode):
                    self.state.lock_mode = state.lock_mode
                elif isinstance(state, StatusAliveFactorySetting):
                    self.state.production_text = f"Produced at: {state.factory_setting.production_date} {state.factory_setting.production_time}"
                    if state.factory_setting.invalid != FactorySettingRev1.INVALID_VALUE_PRODUCTION:
                        self.state.production_text += " (NOT A PRODUCTION UNIT)"

                    self.state.commit_text = f"Commit: {state.factory_setting.git_commit_short}"
                    self.state.serialnumber_text = "HILmate-" + str(state.factory_setting.serialnumber)
                    if not self.state.uc1_mode:
                        self.hilmux.send(CommandGetUc1Mode())
                    if not self.state.uc2_mode:
                        self.hilmux.send(CommandGetUc2Mode())
                    if not self.state.uc1_mode:
                        self.hilmux.send(CommandGetLockMode())
                    self.state.error_text = ""
                elif isinstance(state, StatusBootloaderAliveFactorySetting):
                    self.state.uc1_mode = None
                    self.state.uc2_mode = None
                    self.state.lock_mode = None
                    self.state.error_text = "HILmux is in bootloader!"
                self.root.after(0, lambda: self.apply_state())
        except Exception as ex:
            self.state.error_text = "Error: " + str(ex)
            self.root.after(0, lambda: self.apply_state())


def main():
    parser = set_HILmux.base_arg_parser("HILmux Example GUI")
    args = parser.parse_args()
    discovered_device = set_HILmux.discover_from_args_or_exit(args)
    root = tk.Tk()
    app = DeviceControllerApp(root, discovered_device)
    root.protocol("WM_DELETE_WINDOW", app.on_close)
    root.mainloop()

if __name__ == "__main__":
    main()