"""
ITB-Solutions GmbH HILmux Library

This file/module provides support for easily communicating with HILmux devices.
This file contains:
 - A wrapper around pyserial to easily send Commands and receive Statuses properly.
 - A separately usable abstraction to extract Statuses from a growing byte-stream/buffer
 - Several Discovery-Functions to find HILmux devices on the local system
"""
import hilmux_proto

from dataclasses import dataclass
from serial import Serial
from serial.tools.list_ports import comports
import time
import threading
from typing import Union, Optional, List

from hilmux_proto import CommandGetAliveFactorySetting


#   /------------------------------/
#  /        Communication         /
# /------------------------------/


class TooManyTroublesomeBytes(Exception):
    pass

class StatusDecoder:
    def __init__(self):
        self.buffer = bytearray()
    
    def extend(self, data: Union[bytes, bytearray]):
        self.buffer.extend(data)
    
    def consume(self, count: int):
        self.buffer = self.buffer[count:]

    def try_decode_next(self, may_troublesome_bytes=64) -> Optional[hilmux_proto.Status]:
        size = 1
        skipped = 0
        while size <= len(self.buffer):
            try:
                status = hilmux_proto.Status.decode(self.buffer[:size])
                self.consume(size)
                return status
            except hilmux_proto.InvalidSizeException:
                size += 1
            except Exception as error:
                # If an unexpected error happens:
                #  - Try to skip bytes until it works
                #  - Return error, if exceeded max_bytes_skipped_on_error
                if skipped == may_troublesome_bytes:
                    raise TooManyTroublesomeBytes("Failed to decode Status, even after skipping %d troublesome bytes." % may_troublesome_bytes, error)
                self.consume(1)
                skipped += 1
        return None


HILMUX_SERIAL_BAUDRATE = 57600

class Hilmux:
    """
    Helper for communicating with a HILmux device.
    """

    def __init__(self, port_or_device: Union[Serial, str], **kwargs):
        if isinstance(port_or_device, Serial):
            self.device = port_or_device
        elif isinstance(port_or_device, str):
            self.device = Serial(port_or_device, baudrate=HILMUX_SERIAL_BAUDRATE, **kwargs)
        else:
            raise ValueError("First parameter type to Hilmux is not supported: " + str(type(port_or_device)))
        self.decoder = StatusDecoder()
    
    def send(self, command: hilmux_proto.Command):
        if not isinstance(command, hilmux_proto.Command):
            raise ValueError("Only classes based on hilmux_proto.Command should be sent!")
        self.device.write(command.encode())
        self.device.flush()

    def recv(self) -> Optional[hilmux_proto.Status]:
        status = self.decoder.try_decode_next()
        if status is not None:
            # Old data still contained more Status messages, returning these first before continuing
            return status
        
        # Try to get new bytes
        self.decoder.extend(self.device.read(1))
        if self.device.in_waiting > 0:
            # Read remaining bytes, if more were received
            self.decoder.extend(self.device.read(self.device.in_waiting))

        return self.decoder.try_decode_next()
    
    def close(self):
        self.device.close()
        self.decoder.buffer.clear()



#   /------------------------------/
#  /          Discovery           /
# /------------------------------/


def all_eligable_devices():
    return [ device.device for device in filter(lambda device: device.manufacturer == "FTDI", comports()) ]


@dataclass
class DiscoveredDevice:
    device: str
    serialnumber: str


def discover_serialnumber(device: str, send_command: bool=False) -> Optional[DiscoveredDevice]:
    try:
        hilmux = Hilmux(device, timeout=0)
    except Exception:
        return None # Port is likely invalid
    if send_command:
        hilmux.send(CommandGetAliveFactorySetting())
    start = time.monotonic()
    while time.monotonic() - start <= 3.0:
        try:
            while (status := hilmux.recv()) is not None:
                if isinstance(status, hilmux_proto.StatusAliveFactorySetting):
                    hilmux.close()
                    return DiscoveredDevice(device, str(status.factory_setting.serialnumber))
            time.sleep(0.25)
        except Exception:
            hilmux.close()
            return None # Likely not a hilmux!
    hilmux.close()
    return None


def discover_serialnumbers(devices: List[str]) -> List[DiscoveredDevice]:
    discovered_devices = []
    threads: List[threading.Thread] = []

    def thread_target(device_index: int):
        discovered = discover_serialnumber(devices[device_index])
        if discovered is not None:
            discovered_devices.append(discovered)
    
    for i in range(len(devices)):
        thread = threading.Thread(name=f"HMDisc-{i}", target=thread_target, daemon=True, args=(i,))
        threads.append(thread)
        thread.start()
    
    for thread in threads:
        thread.join()
    
    return discovered_devices


def discover(device: Optional[str], serialnumber: Optional[str]) -> List[DiscoveredDevice]:
    """
    Finds as many hilmux devices as possible, given the filters.
    """
    if serialnumber is not None and serialnumber.upper().startswith("hilmux-"):
        serialnumber = serialnumber[len("HILMUX-"):]

    if device is None and serialnumber is None:
        # no device, no serialnumber specified: Return all devices that are a HILmux
        return discover_serialnumbers(all_eligable_devices())
    elif device is not None and serialnumber is None:
        # device and no serialnumber specified: Only return the device if it is a HILmux (has any serialnumber)
        discovered_device = discover_serialnumber(device)
        return [ discovered_device ] if discovered_device is not None else []
    elif device is None and serialnumber is not None:
        # no device, but serialnumber specified: Search all devices and return only device with matching serialnumber
        return list(filter(lambda discovered: discovered.serialnumber.upper() == serialnumber.upper(), discover_serialnumbers(all_eligable_devices())))
    elif device is not None and serialnumber is not None:
        # device and serialnumber specified: Only return the device, if it has the specified serialnumber
        discovered_device = discover_serialnumber(device)
        return [ discovered_device ] if discovered_device is not None and discovered_device.serialnumber.upper() == serialnumber.upper() else []