"""
ITB-Solutions GmbH HILmux Library

This file/module implements the protocol used by the HILmux (everything except the update procedure itself).
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import date, datetime, time
from enum import Enum
from typing import Union, Optional, Type, List
from struct import pack, unpack


#   /------------------------------/
#  /            Basis             /
# /------------------------------/


MAGIC = b"HMUX"


class InvalidSizeException(Exception):
    """
    This is thrown when too little or too many bytes have been tried to decode.
    When trying to decode, starting from as little bytes as possible, it is recommended to
    retry the decoding until this error doesn't happen anymore (either succeeds or any other error).
    """
    pass


class UnknownEncodingLengthException(Exception):
    """
    Thrown by implementations of Coded.encoded_length() if the size cannot be known without further context (data).
    """
    pass


class Codec(ABC):
    """Used by all other classes to implement encoding and decoding support."""

    @staticmethod
    @abstractmethod
    def encoded_length() -> int:
        """
        How many bytes the encoded data would have.
        Some require more context than available without a class instance.
        These will throw a UnknownEncodingLengthException (currently only packets that contain FactorySetting).
        """
        pass

    def __len__(self) -> int:
        return self.__class__.encoded_length()

    @abstractmethod
    def encode(self) -> bytes:
        """Turn current class instance into bytes to send."""
        pass

    @staticmethod
    @abstractmethod
    def decode(encoded: bytes) -> "Codec":
        """Create an instance of the class from encoded bytes."""
        pass


class Command(Codec):
    """Sent from host to HILmux (request)."""
    ID = None

    @staticmethod
    def encoded_length() -> int:
        return len(MAGIC) + 1
    
    def encode(self) -> bytes:
        return MAGIC + pack("B", self.ID)

    @staticmethod
    def decode(encoded: bytes) -> "Command":
        """
        Use this to decode any Command to the correct type.
        You then use isinstance() or Command.ID to check what specific command you got.
        """
        if len(encoded) < 5:
            raise InvalidSizeException("Encoded data too short")
        if not encoded.startswith(MAGIC):
            raise ValueError("Invalid magic bytes")
        cmd_id = unpack("B", encoded[4:5])[0]
        payload = encoded[5:]
        cmd_class = None
        for maybe_cmd_class in Command.commands():
            if maybe_cmd_class.ID == cmd_id:
                cmd_class = maybe_cmd_class
        if cmd_class is None:
            raise ValueError(f"Unknown command ID: {cmd_id}")
        return cmd_class.decode(payload)
    
    @staticmethod
    def commands() -> List[Type["Command"]]:
        return Command.__subclasses__()


class Status(Codec):
    """Sent from HILmux to host (response)."""
    ID = None

    @staticmethod
    def encoded_length() -> int:
        return len(MAGIC) + 1

    def encode(self) -> bytes:
        return MAGIC + pack("B", self.ID)

    @staticmethod
    def decode(encoded: bytes) -> "Status":
        """
        Use this to decode any Status to the correct type.
        You then use isinstance() or Status.ID to check what specific status you got.
        """
        if len(encoded) < 5:
            raise InvalidSizeException("Encoded data too short")
        if not encoded.startswith(MAGIC):
            raise ValueError("Invalid magic bytes")
        status_id = unpack("B", encoded[4:5])[0]
        payload = encoded[5:]
        status_class = None
        for maybe_status_class in Status.statuses():
            if maybe_status_class.ID == status_id:
                status_class = maybe_status_class
        if status_class is None:
            raise ValueError(f"Unknown status ID: {status_id}")
        return status_class.decode(payload)

    @staticmethod
    def statuses() -> List[Type["Status"]]:
        return Status.__subclasses__()



#   /------------------------------/
#  /        Data types            /
# /------------------------------/


class UcMode(Enum):
    """
    The names for 0x01 and 0x02 might change in future products. For now these are safe to assume though.

    Leds on the device:
     - Disconnected: No LED (off)
     - Xetk: Green LED
     - Lauterbach: Red LED (only one uC can have this active at a time)

    (Lauterbach might be shortened to "Lb" in other products.)
    """
    Disconnected = 0x00
    Xetk = 0x01
    Lauterbach = 0x02


class LockMode(Enum):
    """
    Whether the device is/should be locked.

    When locked, the HILmux will light up a dedicated red LED. When attempting to change uC 1 or uC 2, it will blink for a short time.
    """
    Unlocked = 0x00
    Locked = 0x01


@dataclass
class ProductionDate(Codec):
    """
    Used in FactorySetting. Is 4 bytes and can sometimes also be represented as a uint32_t in other products.
    """
    year: int
    month: int
    day: int

    def __init__(self, year_or_full: Union[int, str, date, "ProductionDate"], month: Optional[int]=None, day: Optional[int]=None):
        if isinstance(year_or_full, int) and  isinstance(month, int) and  isinstance(day, int):
            self.year, self.month, self.day = year_or_full, month, day
        elif isinstance(year_or_full, date):
            self.year, self.month, self.day = year_or_full.year, year_or_full.month, year_or_full.day
        elif isinstance(year_or_full, str):
            d = datetime.strptime(year_or_full, "%Y-%m-%d")
            self.year, self.month, self.day = d.date().year, d.date().month, d.date().day
        elif isinstance(year_or_full, ProductionDate):
            self.year, self.month, self.day = year_or_full.year, year_or_full.month, year_or_full.day
        else:
            raise ValueError("First parameter type to ProductionDate is unsupported or not all 3 parameters are defined: " + str(type(year_or_full)))
        self.date() # Validation

    @staticmethod
    def encoded_length() -> int:
        return 4

    def encode(self) -> bytes:
        return pack("<HBB", self.year, self.month, self.day)
    
    @staticmethod
    def decode(encoded: bytes):
        return ProductionDate(*unpack("<HBB", encoded))
    
    def date(self) -> date:
        return date(self.year, self.month, self.day)
    
    def __str__(self):
        return self.date().strftime("%Y-%m-%d")


@dataclass
class ProductionTime(Codec):
    """
    Used in FactorySetting. Is 4 bytes and can sometimes also be represented as a uint32_t in other products.
    """
    hour: int
    minute: int
    second: int
    
    def __init__(self, hour_or_full: Union[int, str, date, "ProductionTime"], minute: Optional[int]=None, second: Optional[int]=None):
        if isinstance(hour_or_full, int) and  isinstance(minute, int) and  isinstance(second, int):
            self.hour, self.minute, self.second = hour_or_full, minute, second
        elif isinstance(hour_or_full, date):
            self.hour, self.minute, self.second = hour_or_full.hour, hour_or_full.minute, hour_or_full.second
        elif isinstance(hour_or_full, str):
            d = datetime.strptime(hour_or_full, "%H:%M:%S")
            self.hour, self.minute, self.second = d.time().hour, d.time().minute, d.time().second
        elif isinstance(hour_or_full, ProductionTime):
            self.hour, self.minute, self.second = hour_or_full.hour, hour_or_full.minute, hour_or_full.second
        else:
            raise ValueError("First parameter type to ProductionTime is unsupported or not all 3 parameters are defined: " + str(type(hour_or_full)))
        self.time() # Validation

    @staticmethod
    def encoded_length() -> int:
        return 4

    def encode(self) -> bytes:
        return pack("BBBx", self.hour, self.minute, self.second)
    
    @staticmethod
    def decode(encoded: bytes) -> "ProductionTime":
        return ProductionTime(*unpack("BBBx", encoded))
    
    def time(self) -> time:
        return time(self.hour, self.minute, self.second)
    
    def __str__(self):
        return self.time().strftime("%H:%M:%S")


@dataclass
class GitCommitShort:
    data: bytes # 4 bytes

    def __init__(self, source: Union[bytes, bytearray, str]):
        if isinstance(source, bytes) or isinstance(source, bytearray):
            self.data = bytes(source)
        elif isinstance(source, str):
            self.data = bytes.fromhex(source)
        elif isinstance(source, GitCommitShort):
            self.data = source.data
        else:
            raise ValueError("First parameter type to GitCommitShort is unsupported: " + str(type(source)))
        assert len(self.data) == 4

    @staticmethod
    def encoded_length() -> int:
        return 4

    def encode(self) -> bytes:
        return self.data

    @staticmethod
    def decode(encoded: bytes) -> "GitCommitShort":
        return GitCommitShort(encoded)

    def __str__(self):
        return self.data.hex().lower()


@dataclass
class SerialNumber:
    data: bytes # 4 bytes

    def __init__(self, source: Union[bytes, bytearray, str]):
        if isinstance(source, bytes) or isinstance(source, bytearray):
            self.data = bytes(source)
        elif isinstance(source, str):
            self.data = bytes.fromhex(source)
        elif isinstance(source, SerialNumber):
            self.data = source.data
        else:
            raise ValueError("First parameter type to SerialNumber is unsupported: " + str(type(source)))
        assert len(self.data) == 4

    @staticmethod
    def encoded_length() -> int:
        return 4

    def encode(self) -> bytes:
        return self.data

    @staticmethod
    def decode(encoded: bytes) -> "SerialNumber":
        return SerialNumber(encoded)

    def __str__(self):
        return self.data.hex().lower()


class FactorySetting(Codec):
    REVISION = None

    @staticmethod
    def encoded_length() -> int:
        raise UnknownEncodingLengthException("Size of FactorySetting is not known without knowledge of the revision.")
    
    @abstractmethod
    def encode(self) -> bytes:
        return pack("B", self.REVISION)

    @staticmethod
    def decode(encoded: bytes) -> "FactorySetting":
        if len(encoded) < 1:
            raise InvalidSizeException("Encoded data too short for FactorySetting")
        revision = unpack("B", encoded[0:1])[0]
        payload = encoded[1:]
        fs_class = None
        for maybe_fs_class in FactorySetting.revisions():
            if maybe_fs_class.REVISION == revision:
                fs_class = maybe_fs_class
        if fs_class is None:
            raise ValueError(f"Unknown FactorySetting revision: {revision}")
        return fs_class.decode(payload)
    
    @staticmethod
    def revisions() -> List[Type["FactorySetting"]]:
        return FactorySetting.__subclasses__()


@dataclass
class FactorySettingRev1(FactorySetting):
    REVISION = 1
    INVALID_VALUE_PRODUCTION = 0xEE

    invalid: int # u8
    git_commit_short: "GitCommitShort"
    serialnumber: "SerialNumber"
    production_date: "ProductionDate"
    production_time: "ProductionTime"

    def __init__(self, invalid: int, git_commit_short: Union["GitCommitShort", str], serialnumber: Union["SerialNumber", str], production_date: Union["ProductionDate", date, str], production_time: Union["ProductionTime", date, str]):
        self.invalid = invalid
        self.git_commit_short = GitCommitShort(git_commit_short)
        self.serialnumber = SerialNumber(serialnumber)
        self.production_date = ProductionDate(production_date)
        self.production_time = ProductionTime(production_time)

    @staticmethod
    def encoded_length() -> int:
        own_fields_size = 1 + sum(map(lambda cls: cls.encoded_size(), [GitCommitShort, SerialNumber, ProductionDate, ProductionTime]))
        assert own_fields_size == 17
        return super().encoded_length() + own_fields_size

    def encode(self) -> bytes:
        return super().encode() + pack("B", self.invalid) + b"".join(map(lambda val: val.encode(), [self.git_commit_short, self.serialnumber, self.production_date, self.production_time]))

    @staticmethod
    def decode(encoded: bytes):
        if len(encoded) != 17:
            raise InvalidSizeException(f"Encoded data has wrong length for FactorySettingRev1: {len(encoded)} bytes, expected 17")
        return FactorySettingRev1(
            invalid=unpack("B", encoded[0:1])[0],
            git_commit_short=GitCommitShort.decode(encoded[1:5]),
            serialnumber=SerialNumber.decode(encoded[5:9]),
            production_date=ProductionDate.decode(encoded[9:13]),
            production_time=ProductionTime.decode(encoded[13:17]),
        )
    
    def is_production(self):
        return self.invalid == FactorySettingRev1.INVALID_VALUE_PRODUCTION
    
    def __str__(self):
        return "FactorySettingRev1(invalid=0x%s, git_commit_short=%s, serialnumber=%s, production_date=%s, production_time=%s)" % \
            (hex(self.invalid)[2:].upper(), self.git_commit_short, self.serialnumber, self.production_date, self.production_time)

    def __repr__(self):
        return str(self)



#   /------------------------------/
#  /          Commands            /
# /------------------------------/


@dataclass
class CommandSetUc1Mode(Command):
    """
    Change the UcMode for Microchip 1 to the specified one (will result in a StatusUc1Mode response, even if the mode didn't change).
    """
    uc_mode: UcMode
    ID = 0x00

    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length() + 1

    def encode(self) -> bytes:
        return super().encode() + pack("B", self.uc_mode.value)

    @staticmethod
    def decode(payload: bytes) -> "CommandSetUc1Mode":
        if len(payload) != 1:
            raise InvalidSizeException("Invalid payload length for CommandSetUc1Mode")
        uc_mode = UcMode(unpack("B", payload)[0])
        return CommandSetUc1Mode(uc_mode)


@dataclass
class CommandSetUc2Mode(Command):
    """
    Change the UcMode for Microchip 2 to the specified one (will result in a StatusUc2Mode response, even if the mode didn't change).
    """
    uc_mode: UcMode
    ID = 0x01
    
    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length() + 1

    def encode(self) -> bytes:
        return super().encode() + pack("B", self.uc_mode.value)

    @staticmethod
    def decode(payload: bytes) -> "CommandSetUc2Mode":
        if len(payload) != 1:
            raise InvalidSizeException("Invalid payload length for CommandSetUc2Mode")
        uc_mode = UcMode(unpack("B", payload)[0])
        return CommandSetUc2Mode(uc_mode)


@dataclass
class CommandSetLockMode(Command):
    """
    Change the LockMode to the specified one (will result in a StatusLockMode response, even if the mode didn't change).
    """
    lock_mode: LockMode
    ID = 0x02
    
    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length() + 1

    def encode(self) -> bytes:
        return super().encode() + pack("B", self.lock_mode.value)

    @staticmethod
    def decode(payload: bytes) -> "CommandSetLockMode":
        if len(payload) != 1:
            raise InvalidSizeException("Invalid payload length for CommandSetLockMode")
        lock_mode = LockMode(unpack("B", payload)[0])
        return CommandSetLockMode(lock_mode)


@dataclass
class CommandGetUc1Mode(Command):
    """
    Request a StatusUc1Mode response immediately (otherwise only sent when it changes by button press or Command).
    """
    ID = 0x03

    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length()

    @staticmethod
    def decode(payload: bytes) -> "CommandGetUc1Mode":
        if len(payload) != 0:
            raise InvalidSizeException("Invalid payload length for CommandGetUc1Mode")
        return CommandGetUc1Mode()


@dataclass
class CommandGetUc2Mode(Command):
    """
    Request a StatusUc2Mode response immediately (otherwise only sent when it changes by button press or Command).
    """
    ID = 0x04

    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length()

    @staticmethod
    def decode(payload: bytes) -> "CommandGetUc2Mode":
        if len(payload) != 0:
            raise InvalidSizeException("Invalid payload length for CommandGetUc2Mode")
        return CommandGetUc2Mode()


@dataclass
class CommandGetLockMode(Command):
    """
    Request a StatusLockMode response immediately (otherwise only sent when it changes by button press or Command).
    """
    ID = 0x05

    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length()

    @staticmethod
    def decode(payload: bytes) -> "CommandGetLockMode":
        if len(payload) != 0:
            raise InvalidSizeException("Invalid payload length for CommandGetLockMode")
        return CommandGetLockMode()


@dataclass
class CommandGetAliveFactorySetting(Command):
    """
    Request a StatusAliveFactorySetting response immediately (which is already sent periodically).
    """
    ID = 0x06

    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length()

    @staticmethod
    def decode(payload: bytes) -> "CommandGetAliveFactorySetting":
        if len(payload) != 0:
            raise InvalidSizeException("Invalid payload length for CommandGetAliveFactorySetting")
        return CommandGetAliveFactorySetting()


@dataclass
class CommandEnterBootloader(Command):
    """
    Instructs the HILmux to reboot into bootloader. If left for too long, it will start up
    normally again. So it can also be seen as a way to reboot it.
    """
    ID = 0x07

    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length()

    @staticmethod
    def decode(payload: bytes) -> "CommandEnterBootloader":
        if len(payload) != 0:
            raise InvalidSizeException("Invalid payload length for CommandEnterBootloader")
        return CommandEnterBootloader()


@dataclass
class CommandBootloaderEnterLoader(Command):
    """
    Use only when in Bootloader (getting a BootloaderAliveFactorySetting).
    This instructs the bootloader to expect a firmware update, which switches the protocol.
    Until the device is power-cycled, completed its firmware update or otherwise triggered
    a restart, it will not talk in this documented protocol anymore.
    """
    ID = 0xFF

    @staticmethod
    def encoded_length() -> int:
        return Command.encoded_length()

    @staticmethod
    def decode(payload: bytes) -> "CommandBootloaderEnterLoader":
        if len(payload) != 0:
            raise InvalidSizeException("Invalid payload length for CommandBootloaderEnterLoader")
        return CommandBootloaderEnterLoader()



#   /------------------------------/
#  /          Statuses            /
# /------------------------------/


@dataclass
class StatusUc1Mode(Status):
    """
    Reports the current/changed Microchip 2 Mode. Can be triggered via associated Get/Set commands or button presses.
    """
    ID = 0x00
    uc_mode: UcMode

    @staticmethod
    def encoded_length() -> int:
        return Status.encoded_length() + 1

    def encode(self) -> bytes:
        return super().encode() + pack("B", self.uc_mode.value)

    @staticmethod
    def decode(payload: bytes) -> "StatusUc1Mode":
        if len(payload) != 1:
            raise InvalidSizeException("Invalid payload length for StatusUc1Mode")
        uc_mode = UcMode(unpack("B", payload)[0])
        return StatusUc1Mode(uc_mode)


@dataclass
class StatusUc2Mode(Status):
    """
    Reports the current/changed Microchip 2 Mode. Can be triggered via associated Get/Set commands or button presses.
    """
    ID = 0x01
    uc_mode: UcMode

    @staticmethod
    def encoded_length() -> int:
        return Status.encoded_length() + 1

    def encode(self) -> bytes:
        return super().encode() + pack("B", self.uc_mode.value)

    @staticmethod
    def decode(payload: bytes) -> "StatusUc2Mode":
        if len(payload) != 1:
            raise InvalidSizeException("Invalid payload length for StatusUc2Mode")
        uc_mode = UcMode(unpack("B", payload)[0])
        return StatusUc2Mode(uc_mode)


@dataclass
class StatusLockMode(Status):
    """
    Reports the current/changed Lock Mode. Can be triggered via associated Get/Set commands or button presses.
    """
    lock_mode: LockMode
    ID = 0x02

    @staticmethod
    def encoded_length() -> int:
        return Status.encoded_length() + 1

    def encode(self) -> bytes:
        return super().encode() + pack("B", self.lock_mode.value)

    @staticmethod
    def decode(payload: bytes) -> "StatusLockMode":
        if len(payload) != 1:
            raise InvalidSizeException("Invalid payload length for StatusLockMode")
        lock_mode = LockMode(unpack("B", payload)[0])
        return StatusLockMode(lock_mode)


@dataclass
class StatusAliveFactorySetting(Status):
    """
    When the device is running, this gets sent periodically or when requested via the associated command.
    """
    ID = 0x03
    factory_setting: FactorySetting

    @staticmethod
    def encoded_length() -> int:
        raise UnknownEncodingLengthException("Size may not always be the same for StatusAliveFactorySetting")

    def encode(self) -> bytes:
        return super().encode() + self.factory_setting.encode()

    @staticmethod
    def decode(payload: bytes) -> "StatusAliveFactorySetting":
        factory_setting = FactorySetting.decode(payload)
        return StatusAliveFactorySetting(factory_setting)


@dataclass
class StatusBootloaderAliveFactorySetting(Status):
    """
    Same as StatusAliveFactorySetting, however this is sent when in Bootloader.
    Usually this is sent one or more times shortly after powering up or rebooting.
    When this is sent, you may respond with CommandBootloaderEnterLoader to start
    updating the firmware.
    """
    ID = 0xFF
    factory_setting: FactorySetting

    @staticmethod
    def encoded_length() -> int:
        raise UnknownEncodingLengthException("Size may not always be the same for StatusBootloaderAliveFactorySetting")

    def encode(self) -> bytes:
        return super().encode() + self.factory_setting.encode()

    @staticmethod
    def decode(payload: bytes):
        factory_setting = FactorySetting.decode(payload)
        return StatusBootloaderAliveFactorySetting(factory_setting)
