#!/usr/bin/env python3
import hilmux
from hilmux_proto import *
from hilmux import discover, DiscoveredDevice, Hilmux

import argparse
from dataclasses import dataclass
from configparser import ConfigParser
from sys import argv, stderr
from textwrap import dedent
import time


@dataclass
class Config:
    serialnumber: str
    last_known_device: str
    uc1_display_name: str = "uC 1"
    uc2_display_name: str = "uC 2"

    def read_from(config_file: str) -> "Config":
        cp = ConfigParser()
        cp.read(config_file)
        return Config(**dict(cp.items("HILmux")))

    def save_to(self, config_file: str):
        cp = ConfigParser()
        cp["HILmux"] = self.__dict__
        with open(config_file, "w") as f:
            cp.write(f)


config: Config = None


def discover_from_args_or_exit(parsed_args) -> DiscoveredDevice:
    global config

    if parsed_args.config:
        config = Config.read_from(parsed_args.config)
        discovered_device = hilmux.discover_serialnumber(config.last_known_device, True)
        if discovered_device is not None and discovered_device.serialnumber.upper() == config.serialnumber.upper():
            print(f"Confirmed HILmux-{config.serialnumber} to be at {config.last_known_device}")
            return discovered_device

        if discovered_device is None:
            print(f"WARN: Failed to find a HILmux device at {config.last_known_device}!", file=stderr)
        elif discovered_device.serialnumber.upper() != config.serialnumber.upper():
            print(f"WARN: Expected device {config.last_known_device} to be HILmux-{config.serialnumber}, but found HILmux-{discovered_device.serialnumber}!", file=stderr)
        print(f"Searching for HILmux-{config.serialnumber}...")
        discovered_devices = hilmux.discover(None, config.serialnumber)
        if len(discovered_devices) != 1:
            print("Failed to find single HILmux with that serialnumber!", file=stderr)
            exit(1)
        discovered_device = discovered_devices[0]
        if not parsed_args.no_config_update:
            config.last_known_device = discovered_device.device
            config.save_to(parsed_args.config)
        print(f"Found HILmux-{config.serialnumber} at {discovered_device.device}." + ("" if parsed_args.no_config_update else " Updated last_known_device in config (prevent this using --no-config-update)."))
        return discovered_device
    
    discovered_devices = discover(parsed_args.device, parsed_args.serialnumber)
    if len(discovered_devices) == 1:
        print(f"Found HILmux-{discovered_devices[0].serialnumber} at {discovered_devices[0].device}")
        return discovered_devices[0]
    elif len(discovered_devices) == 0:
        print(f"Failed to find any HILmux. Make sure at least one is connected and you used the parameters correctly.\nRun \"{argv[0]} --help\" for more information.", file=stderr)
        exit(1)
    elif len(discovered_devices) > 1:
        print(f"Found more than one device. Please use --config or --device, --serialnumber to make sure only one device is found. Devices: \n" + \
              "\n".join(map(lambda d: f" - {d.device}: {d.serialnumber}"), discovered_devices) + \
              f"\nRun \"{argv[0]} --help\" for more information.", file=stderr)
        exit(1)


def base_arg_parser(description_header) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=dedent(f"""
                            {description_header}

                            To find a HILmux, specify either any combination of --device, --serialnumber,
                            just --config or none of them (to find the only existing HILmux).

                            You can create config files using create_config_file.py.                            
                            """),
                            formatter_class=argparse.RawDescriptionHelpFormatter)

    # Add optional arguments
    parser.add_argument("-c", "--config", type=str, help="Ignore --device, --serialnumber and use values from config (faster)")
    parser.add_argument("-d", "--device", type=str, help="Specify device (e.g. COM3 or /dev/ttyUSB0)")
    parser.add_argument("-s", "--serialnumber", type=str, help="Specify serialnumber (e.g. ABCD0123)")
    parser.add_argument("-U", "--no-config-update", action="store_true", help="Do not allow to update a config file (to fix last_known_device)")
    return parser


def main():
    parser = base_arg_parser("HILmux_set example script")
    
    parser.add_argument("-1", "--uc1", type=str, help="Desired uC1 state", choices=list(map(lambda u: u.name, UcMode)))
    parser.add_argument("-2", "--uc2", type=str, help="Desired uC2 state", choices=list(map(lambda u: u.name, UcMode)))

    # Parse the arguments
    args = parser.parse_args()
    discovered_device = discover_from_args_or_exit(args)

    # Parse desired ucmodes
    desired_uc1_mode: Optional[UcMode] = None
    desired_uc2_mode: Optional[UcMode] = None
    if args.uc1 is not None:
        desired_uc1_mode = next(filter(lambda ucmode: ucmode.name.lower() == args.uc1.lower(), UcMode), None)
    if args.uc2 is not None:
        desired_uc2_mode = next(filter(lambda ucmode: ucmode.name.lower() == args.uc2.lower(), UcMode), None)

    # Open hilmux
    hilmux = Hilmux(discovered_device.device, timeout=0)

    # Query current states
    hilmux.send(CommandGetUc1Mode())
    hilmux.send(CommandGetUc2Mode())
    hilmux.send(CommandGetLockMode())
    current_queried_at = time.monotonic()
    current_uc1_mode: Optional[UcMode] = None
    current_uc2_mode: Optional[UcMode] = None
    current_lock_mode: Optional[LockMode] = None
    while True:
        if current_uc1_mode is not None and current_uc2_mode is not None and current_lock_mode is not None:
            break
        status = hilmux.recv()
        if status is None:
            time.sleep(0.25)
            if time.monotonic() - current_queried_at > 3.0:
                print("Took too long to query current uC and Lock modes!")
                exit(1)
        if isinstance(status, StatusUc1Mode):
            current_uc1_mode = status.uc_mode
        elif isinstance(status, StatusUc2Mode):
            current_uc2_mode = status.uc_mode
        elif isinstance(status, StatusLockMode):
            current_lock_mode = status.lock_mode
    print(f"uC1: {current_uc1_mode.name}, uC2: {current_uc2_mode.name}, Locked: {current_lock_mode.name}")

    if not desired_uc1_mode and not desired_uc2_mode:
        print("Neither --uc1, nor --uc2 specified. Not changing anything.")
        exit(0)
    
    if current_lock_mode == LockMode.Locked:
        print("Device is currently locked!", file=stderr)
        exit(1)
    
    if desired_uc1_mode is not None:
        # Update UC1 Mode
        if desired_uc1_mode == UcMode.Lauterbach and current_uc2_mode == UcMode.Lauterbach:
            print("Can't set uC1 to Lauterbach as uC2 is already using it!", file=stderr)
            exit(1)
        hilmux.send(CommandSetUc1Mode(desired_uc1_mode))
        current_uc1_mode = desired_uc1_mode # Assume it worked
        print(f"Set uC1 to {desired_uc1_mode.name}")
    
    if desired_uc2_mode is not None:
        # Update UC1 Mode
        if desired_uc2_mode == UcMode.Lauterbach and current_uc1_mode == UcMode.Lauterbach:
            print("Can't set Uc2 to Lauterbach as uC1 is already using it!", file=stderr)
            exit(1)
        hilmux.send(CommandSetUc2Mode(desired_uc2_mode))
        current_uc2_mode = desired_uc2_mode # Assume it worked
        print(f"Set uC2 to {desired_uc2_mode.name}")
    
    # Done
    hilmux.close()


if __name__ == "__main__":
    main()