#!/usr/bin/env python3

from hilmux_proto import *

import unittest


class TestProto(unittest.TestCase):

    def test_commands(self):
        expected = [
            (bytes.fromhex("48 4D 55 58    00    00"), CommandSetUc1Mode(UcMode.Disconnected)),
            (bytes.fromhex("48 4D 55 58    01    02"), CommandSetUc2Mode(UcMode.Lauterbach)),
            (bytes.fromhex("48 4D 55 58    02    00"), CommandSetLockMode(LockMode.Unlocked)),
            (bytes.fromhex("48 4D 55 58    03"),       CommandGetUc1Mode()),
            (bytes.fromhex("48 4D 55 58    04"),       CommandGetUc2Mode()),
            (bytes.fromhex("48 4D 55 58    05"),       CommandGetLockMode()),
            (bytes.fromhex("48 4D 55 58    06"),       CommandGetAliveFactorySetting()),
            (bytes.fromhex("48 4D 55 58    07"),       CommandEnterBootloader()),
            (bytes.fromhex("48 4D 55 58    FF"),       CommandBootloaderEnterLoader()),
        ]

        tested_command_classes = set()
        for expected_encoded, expected_command in expected:
            cls_name = expected_command.__class__.__name__
            try:
                self.assertEqual(len(expected_encoded), expected_command.__class__.encoded_length(),
                             f"{cls_name}: Test encoded data doesn't match expected size reported by command")
            except UnknownEncodingLengthException:
                pass # Not all statuses know their final size in advance
            decoded_command = Command.decode(expected_encoded)
            self.assertEqual(decoded_command, expected_command,
                             f"{cls_name}: Decoded test data did not result in same command (expected {expected_command}, got {decoded_command})!")
            re_encoded = decoded_command.encode()
            self.assertEqual(re_encoded, expected_encoded,
                             f"{cls_name}: Re-encoding the decoded command did not result in the same binary data (expected {expected_encoded.hex()}, got {re_encoded.hex()})!")
            tested_command_classes.add(expected_command.__class__)

        self.assertEqual(len(tested_command_classes), len(Command.commands()),
                         "All possible commands should have at least one example here to ensure the implementation has no obvious runtime-errors.")


    def test_statuses(self):
        expected = [
            (bytes.fromhex("48 4D 55 58    00    01"), StatusUc1Mode(UcMode.Xetk)),
            (bytes.fromhex("48 4D 55 58    01    00"), StatusUc2Mode(UcMode.Disconnected)),
            (bytes.fromhex("48 4D 55 58    02    01"), StatusLockMode(LockMode.Locked)),
            (bytes.fromhex("48 4D 55 58    03    01eef00dabfd02020012e807050117303200"),
                StatusAliveFactorySetting(FactorySettingRev1(0xEE, "F00DABFD", "02020012", "2024-05-01", "23:48:50"))),
            (bytes.fromhex("48 4D 55 58    FF    0100f00dabfd01011234e8070a1f083b3b00"),
                StatusBootloaderAliveFactorySetting(FactorySettingRev1(0x00, "F00DABFD", "01011234", "2024-10-31", "08:59:59"))),
        ]

        tested_status_classes = set()
        for expected_encoded, expected_status in expected:
            cls_name = expected_status.__class__.__name__
            try:
                self.assertEqual(len(expected_encoded), expected_status.__class__.encoded_length(),
                             f"{cls_name}: Test encoded data doesn't match expected size reported by status")
            except UnknownEncodingLengthException:
                pass # Not all statuses know their final size in advance
            decoded_status = Status.decode(expected_encoded)
            self.assertEqual(decoded_status, expected_status,
                             f"{cls_name}: Decoded test data did not result in same status (expected {expected_status}, got {decoded_status})!")
            re_encoded = decoded_status.encode()
            self.assertEqual(re_encoded, expected_encoded,
                             f"{cls_name}: Re-encoding the decoded status did not result in the same binary data (expected {expected_encoded.hex()}, got {re_encoded.hex()})!")
            tested_status_classes.add(expected_status.__class__)

        self.assertEqual(len(tested_status_classes), len(Status.statuses()),
                         "All possible commands should have at least one example here to ensure the implementation has no obvious runtime-errors.")


if __name__ == "__main__":
    unittest.main()
